require 'zip'
module Storage
#Class trie - storage
class Storage
  attr_accessor :trie
  def initialize
    @trie=Node.new
  end

  #add str to storage
  def add(str)
    raise ArgumentError if !str.match(/^[[:alpha:][,]]+$/)
    words = str.split(',').reject { |c| c.empty? }.each {|word| @trie.add_word(word)}
  end

  #print all trie words
  def to_s
    @trie.get_subtrie_words([], '').join(",")
  end

  #check contains str in storage
  def contains?(str)
    @trie.contains?(str)
  end

  #find all words starting from str
  def find(str)
    raise ArgumentError if str.length<3
    st = str.dup
    @trie.find([],str,st)
  end

  #load words in storage from file
  def load_from_file(filename)
    lines = IO.readlines(filename)
    lines.map {|line| add(line.strip)}
  end

  #save words from storage to file
  def save_to_file(filename)
    File.open(filename, 'w+') {|f| f.puts(self)}
  end

  #load words in storage from zip
  def load_from_zip(filename)
    Zip::File.foreach(filename) do |entry|
      istream = entry.get_input_stream
      istream.read.split('\n').each do |line|
        add(line)
      end
    end
  end

  #save words from storage to zip
  def save_to_zip(filename)
    Zip::File.open(filename, Zip::File::CREATE) {
        |zipfile|
      zipfile.get_output_stream('storage.txt') { |f| f.puts  @trie.get_subtrie_words([], '').join(',') }
    }
  end

end

#Class Node of trie
class Node
  # subtrie and property of ending word on this node
  attr_accessor :subtrie

  def final!
    @final = true
  end

  def final?
    @final
  end

  def initialize
    @final = false
    @subtrie = {}
  end

  #check contains str starting from current node
  def contains?(str)
    let = str[0]
    if str.empty?
      final?
    elsif @subtrie[let]
      @subtrie[let].contains?(str[1..-1])
    else false
    end
  end

  #add word in trie
  def add_word(word)
    let = word[0]
    if !word.empty?
      if !@subtrie[let]
        @subtrie[let] = Node.new
      end
      @subtrie[let].add_word(word[1..-1])
    else
      final!
    end
  end

  #get all words starting from current node
  def get_subtrie_words(res,str)
    @subtrie.each do |k,node|
      res.push(str+k) if node.final?
      node.get_subtrie_words(res,str+k) if !node.subtrie.empty?
    end
    res
  end

  #find all words starting from current node and str
  def find(res,str,st)
    let = str[0]
    if @subtrie[let]
      if str.length == 1
        res.push(st) if @subtrie[let].final?
        @subtrie[let].get_subtrie_words(res,st)
      else
        @subtrie[let].find(res,str[1..-1],st)
      end
    else
      res
    end
  end
end
end
